package com.codigoparallevar.minicards;

import android.util.Log;

import com.codigoparallevar.minicards.types.functional.Tuple2;
import com.programaker.api.ProgramakerApi;
import com.programaker.api.ProgramakerListeningChannel;
import com.programaker.api.ProgramakerSignalListener;

import org.jetbrains.annotations.NotNull;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class SignalListenerManager implements ProgramakerSignalListener {
    private final ProgramakerApi api;
    private final Map<Tuple2<String, String>, List<ProgramakerSignalListener>> channelToListener = new LinkedHashMap<>();
    private final Map<ProgramakerSignalListener, List<Tuple2<String, String>>> listenerChannels = new LinkedHashMap<>();
    private final Map<Tuple2<String, String>, ProgramakerListeningChannel> idToChannel = new LinkedHashMap<>();
    private long RECONNECT_SLEEP_TIME = 2000; // 2seconds
    private final static String LogTag = "Signal Listener Manager";

    public SignalListenerManager(ProgramakerApi api) {
        this.api = api;
    }

    public void registerSignalListener(String bridgeId, String key, ProgramakerSignalListener listener) {
        Tuple2<String, String> id = new Tuple2<>(bridgeId, key);
        if (!idToChannel.containsKey(id)) {
            // Channel has to be opened
            idToChannel.put(id, this.api.openChannelTo(bridgeId, key, this,
                    () -> {
                        SignalListenerManager.this.onDisconnect(bridgeId, key);
                    }));
        }

        if (!channelToListener.containsKey(id)) {
            channelToListener.put(id, new LinkedList<>());
        }
        List<ProgramakerSignalListener> listeners = channelToListener.get(id);
        listeners.add(listener);

        if (!listenerChannels.containsKey(listener)) {
            listenerChannels.put(listener, new LinkedList<>());
        }
        listenerChannels.get(listener).add(id);
    }

    public void unregisterSignalListener(ProgramakerSignalListener listener) {
        List<Tuple2<String, String>> channels = listenerChannels.get(listener);
        listenerChannels.remove(listener);
        for (Tuple2<String, String> id : channels) {
            List<ProgramakerSignalListener> remainingListeners = channelToListener.get(id);

            remainingListeners.remove(listener);
            if (remainingListeners.size() == 0) {
                ProgramakerListeningChannel channel = idToChannel.get(id);
                idToChannel.remove(id);
                channelToListener.remove(id);

                channel.stop();
            }
        }
    }

    private void onDisconnect(String bridgeId, String key) {
        Tuple2<String, String> id = new Tuple2<>(bridgeId, key);

        // Check that there is still someone listening
        if (!this.channelToListener.containsKey(id)) {
            return;
        }
        if (this.channelToListener.get(id).size() == 0) {
            this.channelToListener.remove(id);
            return;
        }

        // On disconnect disable the connection, wait 2 seconds and retry
        idToChannel.put(id, null);
        
        new Thread(() -> {
            try {
                Thread.sleep(RECONNECT_SLEEP_TIME);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }

            idToChannel.put(id, this.api.openChannelTo(bridgeId, key, this,
                    () -> {
                        SignalListenerManager.this.onDisconnect(bridgeId, key);
                    }));
        }).start();
    }

    @Override
    public void onNewSignal(@NotNull String bridgeId, @NotNull String key, @NotNull HashMap<?, ?> signal) {
        Tuple2<String, String> id = new Tuple2<>(bridgeId, key);

        if (!channelToListener.containsKey(id)) {
            Log.e(LogTag, "Got signal to unlistened channel (bridgeId=" + bridgeId + ",key=" + key + ")");
            return;
        }

        for (ProgramakerSignalListener listener : channelToListener.get(id)) {
            try {
                listener.onNewSignal(bridgeId, key, signal);
            }
            catch (Exception ex) {
                Log.e(LogTag, "Error passing message (bridge=" + bridgeId + ",key" + key
                        + ") to " + listener, ex);
            }
        }
    }
}
