package com.codigoparallevar.minicards.parts.android;

import android.Manifest;
import android.content.Context;
import android.content.pm.PackageManager;
import android.graphics.Color;
import android.graphics.ImageFormat;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.Rect;
import android.hardware.camera2.CameraAccessException;
import android.hardware.camera2.CameraCaptureSession;
import android.hardware.camera2.CameraCharacteristics;
import android.hardware.camera2.CameraDevice;
import android.hardware.camera2.CameraManager;
import android.hardware.camera2.CaptureRequest;
import android.media.Image;
import android.media.ImageReader;
import android.os.Build;
import android.os.Handler;
import android.os.HandlerThread;
import android.util.Log;
import android.view.Surface;

import androidx.annotation.NonNull;
import androidx.annotation.RequiresApi;

import com.codigoparallevar.minicards.CanvasView;
import com.codigoparallevar.minicards.PartInstantiator;
import com.codigoparallevar.minicards.ScrolledCanvas;
import com.codigoparallevar.minicards.parts.connectors.ImageRoundOutputConnector;
import com.codigoparallevar.minicards.parts.style.CardTheme;
import com.codigoparallevar.minicards.types.Moveable;
import com.codigoparallevar.minicards.types.Part;
import com.codigoparallevar.minicards.types.PartConnection;
import com.codigoparallevar.minicards.types.PartGrid;
import com.codigoparallevar.minicards.types.connectors.input.InputConnector;
import com.codigoparallevar.minicards.types.connectors.output.OutputConnector;
import com.codigoparallevar.minicards.types.functional.Tuple2;
import com.codigoparallevar.minicards.types.wireData.ImageSignal;
import com.codigoparallevar.minicards.types.wireData.WireDataType;
import com.codigoparallevar.minicards.utils.Serializations;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

@RequiresApi(api = Build.VERSION_CODES.LOLLIPOP)
public class CameraStreamer implements Part {
    private static final String LogTag = "CameraStreamer";

    private static final int DEFAULT_SIDE_SIZE = 200;
    private final String _id;
    private final PartGrid _partGrid;
    private int _left;
    private int _top;
    private int _right;
    private int _bottom;
    private List<OutputConnector> _outputConnectors;
    private final ImageRoundOutputConnector _imageRoundOutputConnector;
    private final long SLEEP_TIME = 1000;
    private ImageReader imageReader = null;
    private HandlerThread _thread = null;
    private Handler _handler = null;
    private static final int CAPTURE_IMAGE_WIDTH = 256;
    private static final int CAPTURE_IMAGE_HEIGHT = 256;

    private CameraStreamer(String id, PartGrid partGrid, int left, int top, int right, int bottom) {
        _id = id;
        _partGrid = partGrid;
        _left = left;
        _top = top;
        _right = right;
        _bottom = bottom;

        // Create connectors
        _imageRoundOutputConnector = new ImageRoundOutputConnector(
                this,
                _partGrid,
                getOutputConnectorCenterX(), getOutputConnectorCenterY(),
                getOutputConnectRadius());

        _outputConnectors = new LinkedList<>();
        _outputConnectors.add(_imageRoundOutputConnector);
    }


    public CameraStreamer(PartGrid partGrid, int left, int top, int right, int bottom) {
        this(UUID.randomUUID().toString(), partGrid, left, top, right, bottom);
    }


    @Override
    public void moveEnd(int x, int y) {
        final int width = _right - _left;
        final int height = _bottom - _top;

        _left = x - width / 2;
        _right = _left + width;

        _top = y - height / 2;
        _bottom = _top + height;

        _imageRoundOutputConnector.updatePosition(
                getOutputConnectorCenterX(),
                getOutputConnectorCenterY());
    }

    @Override
    public void drop(int x, int y) {
        moveEnd(x, y);
    }

    @Override
    public boolean containsPoint(int x, int y) {
        return (x >= get_left()) && (x <= get_right())
                && (y >= get_top()) && (y <= get_bottom());
    }

    @Override
    public Moveable getMoveable() {
        return this;
    }

    @Override
    public void unlink() {
        pause();

        for (InputConnector input : getInputConnectors()) {
            input.unlink();
        }
    }

    @Override
    public void draw(ScrolledCanvas canvas, boolean devMode) {
        if (devMode){
            drawConnector(canvas);
            drawWires(canvas, devMode);

            Paint paint = new Paint(Paint.ANTI_ALIAS_FLAG);
            paint.setColor(Color.GRAY);

            canvas.drawRect(
                    new Rect(_left, _top,
                            _right, _bottom),
                    paint);


            // Draw a little camera
            Paint iconPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
            iconPaint.setStyle(Paint.Style.STROKE);
            iconPaint.setColor(Color.YELLOW);
            iconPaint.setStrokeWidth(5f);
            iconPaint.setStrokeCap(Paint.Cap.ROUND);
            iconPaint.setStrokeJoin(Paint.Join.ROUND);

            // TODO: Refactor this into something reasonable
            float side = _right - _left;
            float mid = side / 2;
            float quarter = mid / 2;
            float oct = quarter / 2;
            float hex = oct / 2;
            float shex = hex / 2;

            Path path = new Path();
            // Square top left
            path.moveTo(_left + oct + hex + shex, _top + quarter + hex);
            // Square top right
            path.lineTo(_left + mid + hex + shex, _top + quarter + hex);
            // Box-to-projection connection
            path.lineTo(_left + mid + hex + shex, _top + mid);
            // Projection top right
            path.lineTo(_right - quarter - hex + hex + shex, _top + quarter + hex);
            // Projection bottom right
            path.lineTo(_right - quarter - hex + hex + shex, _bottom - quarter - hex);
            // Back to box
            path.lineTo(_left + mid + hex + shex, _top + mid);
            // Box bottom right
            path.lineTo(_left + mid + hex + shex, _bottom - quarter - hex);
            // Box bottom left
            path.lineTo(_left + oct + hex + shex, _bottom - quarter - hex);

            path.close();
            canvas.drawPath(path, iconPaint);
        }
    }

    private void drawConnector(ScrolledCanvas canvas) {
        Paint outerConnectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        outerConnectorPaint.setColor(CardTheme.IMAGE_CONNECTOR_COLOR_OUTER);

        canvas.drawCircle(getOutputConnectorCenterX(),
                getOutputConnectorCenterY(),
                getOutputConnectRadius(),
                outerConnectorPaint);

        Paint innerConnectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        innerConnectorPaint.setColor(CardTheme.IMAGE_CONNECTOR_COLOR_INNER);

        canvas.drawCircle(getOutputConnectorCenterX(),
                getOutputConnectorCenterY(),
                getOutputConnectRadius() / 2,
                innerConnectorPaint);
    }

    private void drawWires(ScrolledCanvas canvas, boolean devMode) {
        for (OutputConnector outputConnector : _outputConnectors){
            outputConnector.drawWires(canvas, devMode);
        }
    }

    @Override
    public int get_left() {
        return _left;
    }

    @Override
    public int get_right() {
        return _right;
    }

    @Override
    public int get_top() {
        return _top;
    }

    @Override
    public int get_bottom() {
        return _bottom;
    }

    @Override
    public void touched() {
        // Just ignore it, as it's a logic component
    }

    @Override
    public List<InputConnector> getInputConnectors() {
        return Collections.emptyList();
    }

    @Override
    public List<OutputConnector> getOutputConnectors() {
        return _outputConnectors;
    }

    @Override
    public JSONObject serialize() throws JSONException {
        JSONObject serialized = new JSONObject();

        serialized.put("id", _id);
        serialized.put("left", _left);
        serialized.put("top", _top);
        serialized.put("right", _right);
        serialized.put("bottom", _bottom);
        serialized.put("on_signal_output_connector",
                Serializations.serialize(serializeConnectionEndpoints()));

        return serialized;
    }

    public static Tuple2<Part, List<PartConnection>> deserialize(PartGrid partGrid, JSONObject data) throws JSONException {
        String id = data.getString("id");
        int left = data.getInt("left");
        int top = data.getInt("top");
        int right = data.getInt("right");
        int bottom = data.getInt("bottom");

        CameraStreamer cameraStreamer = new CameraStreamer(id, partGrid, left, top, right, bottom);

        List<PartConnection> connections = new LinkedList<>();

        JSONArray connectorOuts = data.getJSONArray("on_signal_output_connector");
        for (int i = 0; i < connectorOuts.length(); i++){
            connections.add(PartConnection.deserialize(
                    cameraStreamer._imageRoundOutputConnector,
                    connectorOuts.getJSONObject(i)));
        }

        return new Tuple2<Part, List<PartConnection>>(cameraStreamer, connections);
    }

    private List<Map<String, String>> serializeConnectionEndpoints() {
        List<Map<String, String>> serializedData = new LinkedList<>();

        for (Tuple2<String, String> endpoint : _imageRoundOutputConnector.getConnectionEndpoints()){
            serializedData.add(PartConnection.serialize(endpoint.item1, endpoint.item2));
        }

        return serializedData;
    }

    @Override
    public void send(InputConnector roundInputConnector, WireDataType signal) {
        // @TODO: REMOVE THE NEED FOR THIS
    }

    private void onNewImage(Image img) {
        _imageRoundOutputConnector.send(new ImageSignal(img));
    }

    @Override
    public String get_id() {
        return _id;
    }

    @Override
    public InputConnector getConnectorWithId(String inputConnectorId) {
        return null;
    }

    @Override
    public String getConnectorId(InputConnector inputConnector) {
        return null;
    }

    @Override
    public void resume()  {
        if (!(this._partGrid instanceof CanvasView)) {
            return;
        }

        CanvasView view = ((CanvasView) this._partGrid);
        Context ctx = view.getContext();

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            if (ctx.checkSelfPermission(Manifest.permission.CAMERA) != PackageManager.PERMISSION_GRANTED) {
                view.requestPermission(Manifest.permission.CAMERA, () -> { CameraStreamer.this.resume(); });
                return;
            }
        }

        this.imageReader = ImageReader.newInstance(CAPTURE_IMAGE_WIDTH, CAPTURE_IMAGE_HEIGHT, ImageFormat.JPEG, 30);
        this.imageReader.setOnImageAvailableListener((ImageReader.OnImageAvailableListener) newImageReader -> {
            Image latestImage = newImageReader.acquireLatestImage();

            if (latestImage == null) {
                return;
            }

            CameraStreamer.this.onNewImage(latestImage);
            latestImage.close();
        }, new Handler());

        CameraManager cm = (CameraManager) ctx.getSystemService(Context.CAMERA_SERVICE);
        try {
            String[] cameraList = cm.getCameraIdList();
            for (String cd : cameraList) {
                //get camera characteristics
                CameraCharacteristics mCameraCharacteristics = cm.getCameraCharacteristics(cd);

                //check if the camera is in the back - if not, continue to next
                if (mCameraCharacteristics.get(CameraCharacteristics.LENS_FACING) != CameraCharacteristics.LENS_FACING_BACK) {
                    continue;
                }

                if (this._thread == null) {
                    this._thread = new HandlerThread("mCameraHandlerThread");
                    this._thread.start();
                }
                if (this._handler == null) {
                    this._handler = new Handler(this._thread.getLooper());
                }

                cm.openCamera(cd, new CameraDevice.StateCallback() {
                    @Override
                    public void onOpened(@NonNull CameraDevice camera) {
                        //make list of surfaces to give to camera
                        List<Surface> surfaceList = new ArrayList<>();
                        Surface surface = CameraStreamer.this.imageReader.getSurface();
                        surfaceList.add(surface);

                        try {
                            camera.createCaptureSession(surfaceList, new CameraCaptureSession.StateCallback() {
                                @Override
                                public void onConfigured(@NonNull CameraCaptureSession session) {
                                    try {
                                        CaptureRequest.Builder requestBuilder = session.getDevice().createCaptureRequest(CameraDevice.TEMPLATE_PREVIEW);
                                        requestBuilder.set(CaptureRequest.JPEG_ORIENTATION, mCameraCharacteristics.get(CameraCharacteristics.SENSOR_ORIENTATION));
                                        requestBuilder.addTarget(surface);
                                        //set to null - image data will be produced but will not receive metadata
                                        session.setRepeatingRequest(requestBuilder.build(), null, CameraStreamer.this._handler);
                                    } catch (CameraAccessException e) {
                                        Log.e(LogTag, "createCaptureSession threw CameraAccessException.", e);
                                    }
                                }

                                @Override
                                public void onConfigureFailed(@NonNull CameraCaptureSession session) {
                                    Log.i(LogTag, "Error on camera configuration");
                                }
                            }, CameraStreamer.this._handler);
                        } catch (CameraAccessException e) {
                            Log.e(LogTag, "createCaptureSession threw CameraAccessException.", e);
                        }
                    }

                    @Override
                    public void onDisconnected(@NonNull CameraDevice camera) {
                        Log.i(LogTag, "Camera disconnected");
                    }

                    @Override
                    public void onError(@NonNull CameraDevice camera, int error) {
                        Log.i(LogTag, "Error on camera opening: " + error);
                    }
                }, this._handler);
                break;
            }

        } catch (CameraAccessException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void pause() {
        this._handler = null;
        if (this._thread != null) {
            this._thread.interrupt();
            this._thread = null;
        }
        if (this.imageReader != null) {
            this.imageReader.close();
            this.imageReader = null;
        }
    }

    @Override
    public Object query(Object lastValue) {
        return null; // No relevant value (maybe time?)
    }


    private int getOutputConnectorCenterX() {
        return (_left + _right) / 2;
    }

    private int getOutputConnectorCenterY() {
        return _bottom;
    }

    private int getOutputConnectRadius() {
        return (_right - _left) / 2;
    }

    public static PartInstantiator getInstantiator() {
        final int halfSideSize = DEFAULT_SIDE_SIZE / 2;
        return new PartInstantiator() {
            @Override
            protected Part instantiate(PartGrid grid, Tuple2<Integer, Integer> center) {
                return new CameraStreamer(grid,
                        center.item1 - halfSideSize, center.item2 - halfSideSize,
                        center.item1 + halfSideSize, center.item2 + halfSideSize);
            }
        };
    }
}
