package com.codigoparallevar.minicards.parts.buttons;

import android.graphics.Color;
import android.graphics.Paint;
import android.util.Log;

import com.codigoparallevar.minicards.PartInstantiator;
import com.codigoparallevar.minicards.ScrolledCanvas;
import com.codigoparallevar.minicards.parts.connectors.SignalRoundOutputConnector;
import com.codigoparallevar.minicards.parts.style.CardTheme;
import com.codigoparallevar.minicards.types.Moveable;
import com.codigoparallevar.minicards.types.Part;
import com.codigoparallevar.minicards.types.PartConnection;
import com.codigoparallevar.minicards.types.PartGrid;
import com.codigoparallevar.minicards.types.functional.Tuple2;
import com.codigoparallevar.minicards.types.connectors.input.InputConnector;
import com.codigoparallevar.minicards.types.connectors.output.OutputConnector;
import com.codigoparallevar.minicards.types.wireData.Signal;
import com.codigoparallevar.minicards.types.wireData.WireDataType;
import com.codigoparallevar.minicards.utils.Serializations;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

public class RoundButton implements Part {
    private final String _id;
    private final PartGrid _partGrid;
    private int _xCenter;
    private int _yCenter;
    private final int _innerRadius;
    private final int _outerRadius;
    private final int _outerRadiusThickness = 10;
    private final int _pathRunWay = 200;
    private List<OutputConnector> _outputConnectors;
    private final SignalRoundOutputConnector _pressedOutputConnector;
    private final static int DEFAULT_INNER_RADIUS = 80;
    private final static int DEFAULT_OUTER_RADIUS = 100;


    private RoundButton(String id, PartGrid partGrid, int xCenter, int yCenter, int innerRadius, int outerRadius) {
        _id = id;

        _partGrid = partGrid;
        _xCenter = xCenter;
        _yCenter = yCenter;
        _innerRadius = innerRadius;
        _outerRadius = outerRadius;

        // Create connectors
        _pressedOutputConnector = new SignalRoundOutputConnector(
                this,
                _partGrid,
                getOutputConnectorCenterX(), getOutputConnectorCenterY(),
                getOutputConnectRadius());

        _outputConnectors = new LinkedList<>();
        _outputConnectors.add(_pressedOutputConnector);
    }

    public RoundButton(PartGrid partGrid, int xCenter, int yCenter, int innerRadius, int outerRadius) {
        this(UUID.randomUUID().toString(), partGrid, xCenter, yCenter, innerRadius, outerRadius);
    }

    @Override
    public int get_left() {
        return _xCenter - _outerRadius;
    }

    @Override
    public int get_right() {
        return _xCenter + _outerRadius;
    }

    @Override
    public int get_top() {
        return _yCenter - _outerRadius;
    }

    @Override
    public int get_bottom() {
        return _yCenter + _outerRadius;
    }

    @Override
    public void draw(ScrolledCanvas canvas, boolean devMode) {
        Paint foregroundPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        foregroundPaint.setColor(Color.WHITE);

        Paint backgroundPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        backgroundPaint.setColor(Color.BLACK);

        if (devMode){
            drawConnector(canvas);
            drawWires(canvas, devMode);
        }

        canvas.drawCircle(_xCenter, _yCenter, _outerRadius, foregroundPaint);
        canvas.drawCircle(_xCenter, _yCenter, _outerRadius - _outerRadiusThickness, backgroundPaint);
        canvas.drawCircle(_xCenter, _yCenter, _innerRadius, foregroundPaint);
    }

    private void drawConnector(ScrolledCanvas canvas) {
        Paint outerConnectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        outerConnectorPaint.setColor(CardTheme.PULSE_CONNECTOR_COLOR_OUTER);

        canvas.drawCircle(getOutputConnectorCenterX(), getOutputConnectorCenterY(),
                getOutputConnectRadius(),
                outerConnectorPaint);

        Paint innerConnectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        innerConnectorPaint.setColor(CardTheme.PULSE_CONNECTOR_COLOR_INNER);

        canvas.drawCircle(getOutputConnectorCenterX(), getOutputConnectorCenterY(),
                getOutputConnectRadius() / 2,
                innerConnectorPaint);
    }

    private int getOutputConnectorCenterX() {
        return _xCenter;
    }

    private int getOutputConnectorCenterY() {
        return _yCenter +  _outerRadius;
    }

    private int getOutputConnectRadius() {
        return _innerRadius;
    }

    private void drawWires(ScrolledCanvas canvas, boolean devMode) {
        for (OutputConnector outputConnector : _outputConnectors){
            outputConnector.drawWires(canvas, devMode);
        }
    }

    @Override
    public void moveEnd(int x, int y) {
        _xCenter = x;
        _yCenter = y;

        // Move connectors too
        _pressedOutputConnector.updatePosition(
                getOutputConnectorCenterX(),
                getOutputConnectorCenterY());
    }

    @Override
    public void drop(int x, int y) {
        moveEnd(x, y);
    }

    @Override
    public void touched() {
        Log.d("RoundButton", "Round button touched");

        _pressedOutputConnector.send(new Signal());
    }

    @Override
    public List<InputConnector> getInputConnectors() {
        return Collections.emptyList();
    }

    @Override
    public List<OutputConnector> getOutputConnectors() {
        return _outputConnectors;
    }

    @Override
    public JSONObject serialize() throws JSONException {
        JSONObject serialized = new JSONObject();

        serialized.put("id", _id);
        serialized.put("x_center", _xCenter);
        serialized.put("y_center", _yCenter);
        serialized.put("inner_radius", _innerRadius);
        serialized.put("outer_radius", _outerRadius);
        serialized.put("outer_radius_thickness", _outerRadiusThickness);
        serialized.put("on_pressed_output_connector",
                Serializations.serialize(serializeConnectionEndpoints()));

        return serialized;
    }

    private List<Map<String, String>> serializeConnectionEndpoints() {
        List<Map<String, String>> serializedData = new LinkedList<>();

        for (Tuple2<String, String> endpoint : _pressedOutputConnector.getConnectionEndpoints()){
            serializedData.add(PartConnection.serialize(endpoint.item1, endpoint.item2));
        }

        return serializedData;
    }

    @Override
    public void send(InputConnector roundInputConnector, WireDataType signal) {
        // @TODO: REMOVE THE NEED FOR THIS
    }

    @Override
    public String get_id() {
        return _id;
    }

    @Override
    public InputConnector getConnectorWithId(String inputConnectorId) {
        return null;
    }

    @Override
    public String getConnectorId(InputConnector inputConnector) {
        return null;
    }

    @Override
    public void resume() {

    }

    @Override
    public void pause() {

    }

    @Override
    public Object query(Object lastValue) {
        return null; // Only pulse output, so no relevant data
    }

    public SignalRoundOutputConnector getPressedOutputConnector(){
        return _pressedOutputConnector;
    }

    public static Tuple2<Part, List<PartConnection>> deserialize(PartGrid partGrid, JSONObject data) throws JSONException {
        String id = data.getString("id");
        int xCenter = data.getInt("x_center");
        int yCenter = data.getInt("y_center");

        int innerRadius = data.getInt("inner_radius");
        int outerRadius = data.getInt("outer_radius");

        RoundButton button = new RoundButton(id, partGrid, xCenter, yCenter, innerRadius, outerRadius);

        List<PartConnection> connections = new LinkedList<>();

        JSONArray connectorOuts = data.getJSONArray("on_pressed_output_connector");
        for (int i = 0; i < connectorOuts.length(); i++){
            connections.add(PartConnection.deserialize(
                    button._pressedOutputConnector,
                    connectorOuts.getJSONObject(i)));
        }

        return new Tuple2<Part, List<PartConnection>>(button, connections);
    }

    @Override
    public boolean containsPoint(int x, int y) {
        return ((Math.abs(x - _xCenter) <= _outerRadius)
                && (Math.abs(y - _yCenter) <= _outerRadius));
    }

    @Override
    public Moveable getMoveable() {
        return this;
    }


    @Override
    public void unlink() {
        for (InputConnector input : getInputConnectors()) {
            input.unlink();
        }
    }

    public static PartInstantiator getInstantiator() {
        return new PartInstantiator() {
            @Override
            protected Part instantiate(PartGrid grid, Tuple2<Integer, Integer> center) {
                return new RoundButton(grid,
                        center.item1, center.item2,
                        DEFAULT_INNER_RADIUS, DEFAULT_OUTER_RADIUS);
            }
        };
    }
}
