package com.codigoparallevar.minicards.parts.connectors;

import android.util.Log;

import com.codigoparallevar.minicards.ScrolledCanvas;
import com.codigoparallevar.minicards.types.Drawable;
import com.codigoparallevar.minicards.types.Part;
import com.codigoparallevar.minicards.types.PartGrid;
import com.codigoparallevar.minicards.types.connectors.Wiring.AnyWire;
import com.codigoparallevar.minicards.types.connectors.input.AnyInputConnector;
import com.codigoparallevar.minicards.types.connectors.input.InputConnector;
import com.codigoparallevar.minicards.types.connectors.output.AnyOutputConnector;
import com.codigoparallevar.minicards.types.functional.Tuple2;
import com.codigoparallevar.minicards.types.wireData.AnySignal;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

public class AnyRoundOutputConnector implements Drawable, AnyOutputConnector,
        RoundOutputConnector<AnySignal, AnyInputConnector, AnyWire> {
    private PartGrid _partGrid;
    private int _centerX;
    private int _centerY;
    private final int _radius;
    private final Part _part;
    private AnyWire _currentWire = null;
    private final List<AnyWire> _wires;
    private final HashSet<InputConnector> _connections;

    public AnyRoundOutputConnector(Part part, PartGrid partGrid, int centerX, int centerY, int radius) {
        _part = part;
        _partGrid = partGrid;
        _centerX = centerX;
        _centerY = centerY;
        _radius = radius;
        _wires = new LinkedList<>();
        _connections = new HashSet<>();
    }

    @Override
    public boolean containsPoint(int x, int y) {
        return ((Math.abs(x - _centerX) <= _radius)
                && (Math.abs(y - _centerY) <= _radius));
    }


    @Override
    public AnyWire getMoveable() {
        if (_currentWire == null) {
            startWire();
        }

        return _currentWire;
    }

    @Override
    public void unlink() {
        while (_wires.size() > 0) {
            _wires.get(0).unlink();
        }
    }

    private void startWire() {
        _currentWire = new AnyWire(this, _centerX, _centerY);
    }

    public void drop(AnyWire wire, AnyInputConnector resultPoint) {
        if (wire != _currentWire) {
            Log.w("RoundOutputConnector",
                    "Asked to drop non matching wire "
                            + "(expected " + _currentWire + ", got " + wire + ")");
            return ;
        }

        _currentWire = null;

        Log.d("RoundOutputConnector", "Dropped wire on " + resultPoint);

        // Not connected
        if (resultPoint == null) {
            return;
        }

        // Already connected
        if (_connections.contains(resultPoint)) {
            return;
        }
        _connections.add(resultPoint);

        wire.attachTo(resultPoint, this);
        _wires.add(wire);
    }

    @Override
    public void drop(AnyWire wire) {

        if (wire != _currentWire) {
            Log.w("RoundOutputConnector",
                    "Asked to drop non matching wire "
                            + "(expected " + _currentWire + ", got " + wire + ")");
            return;
        }

        AnyInputConnector resultPoint = _partGrid.getAnyInputConnectorOn(
                wire.getXEnd(), wire.getYEnd());

        drop(wire, resultPoint);
    }

    @Override
    public void wireUnlinked(AnyWire wire) {
        _wires.remove(wire);
        _connections.remove(wire.getAttachedTo());
    }

    @Override
    public void drawWires(ScrolledCanvas canvas, boolean devMode) {
        for (AnyWire wire : _wires) {
            wire.draw(canvas, devMode);
        }

        if (_currentWire != null) {
            _currentWire.draw(canvas, devMode);
        }
    }

    @Override
    public void updatePosition(int x, int y) {
        _centerX = x;
        _centerY = y;

        for (AnyWire wire : _wires){
            wire.moveStart(x, y);
        }
    }

    @Override
    public List<Tuple2<String, String>> getConnectionEndpoints() {
        List<Tuple2<String, String>> endpointIds = new LinkedList<>();

        for (AnyWire wire : _wires) {
            InputConnector inputConnector = wire.getAttachedTo();
            Part endPart = inputConnector.getPart();
            endpointIds.add(new Tuple2<>(inputConnector.getId(), endPart.get_id()));
        }

        return endpointIds;
    }

    @Override
    public void connectTo(AnyInputConnector inputConnector) {
        if (_connections.contains(inputConnector)) {
            return;
        }

        _connections.add(inputConnector);

        AnyWire wire = new AnyWire(this, _centerX, _centerY);
        wire.attachTo(inputConnector, this);
        _wires.add(wire);
    }

    @Override
    public void draw(ScrolledCanvas canvas, boolean devMode) {
        // TODO: Complete this part
    }

    public void send(AnySignal signal) {
        for (AnyWire wire : _wires){
            wire.send(signal);
        }
    }

    @Override
    public Object query(Object lastValue) {
        return _part.query(lastValue);
    }

    @Override
    public List<AnyWire> getWires() {
        return _wires;
    }

    public int getX() {
        return _centerX;
    }

    public int getY() {
        return _centerY;
    }

    public float getRadius() {
        return _radius;
    }
}
