package com.codigoparallevar.minicards.parts.viewers;

import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Color;
import android.graphics.ImageFormat;
import android.graphics.Paint;
import android.graphics.Rect;
import android.graphics.RectF;
import android.media.Image;
import android.os.Build;
import android.util.Log;

import com.codigoparallevar.minicards.PartInstantiator;
import com.codigoparallevar.minicards.ScrolledCanvas;
import com.codigoparallevar.minicards.parts.connectors.ImageRoundInputConnector;
import com.codigoparallevar.minicards.parts.style.CardTheme;
import com.codigoparallevar.minicards.types.Moveable;
import com.codigoparallevar.minicards.types.Part;
import com.codigoparallevar.minicards.types.PartGrid;
import com.codigoparallevar.minicards.types.connectors.input.ImageInputConnector;
import com.codigoparallevar.minicards.types.connectors.input.InputConnector;
import com.codigoparallevar.minicards.types.connectors.output.OutputConnector;
import com.codigoparallevar.minicards.types.functional.Tuple2;
import com.codigoparallevar.minicards.types.wireData.ImageSignal;
import com.codigoparallevar.minicards.types.wireData.WireDataType;

import org.json.JSONException;
import org.json.JSONObject;

import java.nio.ByteBuffer;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;

public class ImageFrame implements Part {
    private static final String LogTag = "ImageFrame";
    private static final int DEFAULT_SIDE_SIZE = 400;
    private static final String IMAGE_INPUT_CONNECTOR_ID = "image_input_connector";

    private final String _id;
    private final PartGrid _partGrid;
    private int _left;
    private int _top;
    private int _right;
    private int _bottom;
    private List<InputConnector> inputConnectors;
    private ImageInputConnector _imageInputConnector;
    private Bitmap lastImage;

    public ImageFrame(String id, PartGrid partGrid, int left, int top, int right, int bottom) {
        _id = id;
        _partGrid = partGrid;
        _left = left;
        _top = top;
        _right = right;
        _bottom = bottom;

        _imageInputConnector = new ImageRoundInputConnector(
                this,
                getInputConnectorCenterX(),
                getInputConnectorCenterY(),
                getInputConnectRadius());
        inputConnectors = new LinkedList<>();
        inputConnectors.add(_imageInputConnector);
    }

    public  ImageFrame(PartGrid partGrid, int left, int top, int right, int bottom) {
        this(UUID.randomUUID().toString(), partGrid, left, top, right, bottom);
    }

    @Override
    public int get_left() {
        return _left;
    }

    @Override
    public int get_right() {
        return _right;
    }

    @Override
    public int get_top() {
        return _top;
    }

    @Override
    public int get_bottom() {
        return _bottom;
    }

    @Override
    public void draw(ScrolledCanvas canvas, boolean devMode) {
        if (devMode){
            drawConnector(canvas);
            // drawWires(canvas, devMode);
        }

        Rect frameRect = new Rect(_left, _top, _right, _bottom);
        if (this.lastImage == null) {
            Paint backdropPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
            backdropPaint.setColor(Color.parseColor("#444444"));

            canvas.drawRoundRect(new RectF(frameRect), 50, 50, backdropPaint);

            Paint crossPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
            crossPaint.setColor(Color.WHITE);

            // Top
            canvas.drawLine(_left, _top, _right, _top, crossPaint);
            // Bottom
            canvas.drawLine(_left, _bottom, _right, _bottom, crossPaint);
            // Left
            canvas.drawLine(_left, _top, _left, _bottom, crossPaint);
            // Right
            canvas.drawLine(_right, _top, _right, _bottom, crossPaint);
            // Cross, top-left, bottom-right
            canvas.drawLine(_left, _top, _right, _bottom, crossPaint);
            // Cross, top-right, bottom-left
            canvas.drawLine(_right, _top, _left, _bottom, crossPaint);
        }
        else {
            canvas.drawBitmap(this.lastImage, frameRect);
        }
    }

    private void drawConnector(ScrolledCanvas canvas) {
        Paint outerInputConnectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        outerInputConnectorPaint.setColor(CardTheme.IMAGE_CONNECTOR_COLOR_OUTER);

        canvas.drawCircle(
                getInputConnectorCenterX(),
                getInputConnectorCenterY(),
                getInputConnectRadius(),
                outerInputConnectorPaint);

        Paint innerInputConnectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        innerInputConnectorPaint.setColor(CardTheme.IMAGE_CONNECTOR_COLOR_INNER);

        canvas.drawCircle(
                getInputConnectorCenterX(),
                getInputConnectorCenterY(),
                getInputConnectRadius() / 2,
                innerInputConnectorPaint);
    }

    @Override
    public void moveEnd(int x, int y) {
        final int width = _right - _left;
        final int height = _bottom - _top;

        _left = x - width / 2;
        _right = _left + width;

        _top = y - height / 2;
        _bottom = _top + height;

        _imageInputConnector.updatePosition(
                getInputConnectorCenterX(),
                getInputConnectorCenterY());
    }

    @Override
    public void drop(int x, int y) {
        moveEnd(x, y);
    }

    @Override
    public void touched() {
        Log.d(LogTag, "Touched");
    }

    @Override
    public List<InputConnector> getInputConnectors() {
        return inputConnectors;
    }

    @Override
    public List<OutputConnector> getOutputConnectors() {
        return Collections.emptyList();
    }

    @Override
    public JSONObject serialize() throws JSONException {
        JSONObject serialized = new JSONObject();

        serialized.put("id", _id);
        serialized.put("left", _left);
        serialized.put("top", _top);
        serialized.put("right", _right);
        serialized.put("bottom", _bottom);

        return serialized;
    }

    @Override
    public void send(InputConnector roundInputConnector, WireDataType signal) {
        if (!(signal instanceof ImageSignal)) {
            Log.e(LogTag, "Mismatched type, expected {ImageSignal}, found {" + signal + "}" );
        }
        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.KITKAT) {
            // No way to get the format?
            return;
        }

        ImageSignal imgSignal = (ImageSignal) signal;
        Image image = imgSignal.get();
        if (image.getFormat() == ImageFormat.JPEG) {
            Image.Plane[] planes = image.getPlanes();
            if (planes.length != 1) {
                // TODO: Handle this case (what would it mean?)
                return ;
            }

            ByteBuffer jpegBuffer = planes[0].getBuffer();
            byte[] jpegData = new byte[jpegBuffer.remaining()];
            jpegBuffer.get(jpegData);

            Bitmap bitmap = BitmapFactory.decodeByteArray(jpegData, 0, jpegData.length);
            this.lastImage = bitmap;
            this._partGrid.update(); // Maybe this can be avoided? Directly copying the data on the canvas?
        }
    }

    @Override
    public String get_id() {
        return _id;
    }

    @Override
    public InputConnector getConnectorWithId(String inputConnectorId) {
        switch (inputConnectorId){
            case ImageFrame.IMAGE_INPUT_CONNECTOR_ID:
                return _imageInputConnector;

            default:
                return null;
        }
    }

    @Override
    public String getConnectorId(InputConnector inputConnector) {
        if (inputConnector == _imageInputConnector){
            return ImageFrame.IMAGE_INPUT_CONNECTOR_ID;
        }

        return null;
    }
    @Override
    public void resume() {

    }

    @Override
    public void pause() {

    }

    @Override
    public Object query(Object lastValue) {
        return null;
    }

    public static Part deserialize(PartGrid partGrid, JSONObject data) throws JSONException {
        String id = data.getString("id");
        int left = data.getInt("left");
        int top = data.getInt("top");
        int right = data.getInt("right");
        int bottom = data.getInt("bottom");

        return new ImageFrame(id, partGrid, left, top, right, bottom);
    }

    @Override
    public boolean containsPoint(int x, int y) {
        return (x >= get_left()) && (x <= get_right())
                && (y >= get_top()) && (y <= get_bottom());
    }

    @Override
    public Moveable getMoveable() {
        return this;
    }

    @Override
    public void unlink() {
        for (InputConnector input : getInputConnectors()) {
            input.unlink();
        }
    }

    public int getInputConnectorCenterX() {
        return (get_left() + get_right()) / 2;
    }

    private int getInputConnectRadius() {
        return (get_right() - get_left()) / 4;
    }

    public int getInputConnectorCenterY() {
        return get_top();
    }

    public static PartInstantiator getInstantiator() {
        final int halfSideSize = DEFAULT_SIDE_SIZE / 2;
        return new PartInstantiator() {
            @Override
            protected Part instantiate(PartGrid grid, Tuple2<Integer, Integer> center) {
                return new ImageFrame(grid,
                        center.item1 - halfSideSize, center.item2 - halfSideSize,
                        center.item1 + halfSideSize, center.item2 + halfSideSize);
            }
        };
    }
}
