package com.codigoparallevar.minicards.types.connectors.Wiring;

import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.RectF;

import com.codigoparallevar.minicards.ScrolledCanvas;
import com.codigoparallevar.minicards.types.Drawable;
import com.codigoparallevar.minicards.types.Dropper;
import com.codigoparallevar.minicards.types.Moveable;
import com.codigoparallevar.minicards.types.Selectable;
import com.codigoparallevar.minicards.types.connectors.input.InputConnector;
import com.codigoparallevar.minicards.types.connectors.output.OutputConnector;
import com.codigoparallevar.minicards.types.wireData.WireDataType;

public class Wire<T extends WireDataType, InputConnectorType extends InputConnector<T, InputConnectorType>>
        implements Moveable, Drawable, Selectable {

    private final double MAX_DISTANCE_TO_CONTAINS = 15;

    private final Dropper _dropper;
    private int _xinit;
    private int _yinit;
    private int _xend;
    private int _yend;
    private Path path = null;
    private final static int _pathRunWay = 100;
    protected InputConnectorType _attachedTo = null;
    protected OutputConnector _attachedFrom = null;

    public Wire(Dropper dropper, int xInit, int yInit) {
        _dropper = dropper;
        _xinit = xInit;
        _yinit = yInit;

        _xend = xInit;
        _yend = yInit;
    }

    public void moveStart(int x, int y){
        _xinit = x;
        _yinit = y;
        invalidatePath();
    }

    @Override
    public void moveEnd(int x, int y) {
        _xend = x;
        _yend = y;
        invalidatePath(); // Invalidate current path
    }

    @Override
    public void drop(int x, int y) {
        _dropper.drop(this);
    }

    @Override
    public void draw(ScrolledCanvas canvas, boolean devMode) {
        Paint pathPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        pathPaint.setColor(Color.parseColor("#44FF44"));
        pathPaint.setStrokeWidth(10.0f);
        pathPaint.setStrokeCap(Paint.Cap.ROUND);
        pathPaint.setStyle(Paint.Style.STROKE);

        canvas.drawPath(getPath(), pathPaint);
    }



    public int getXEnd() {
        return _xend;
    }

    public int getYEnd() {
        return _yend;
    }

    public void attachTo(InputConnectorType resultPoint, OutputConnector output) {
        _attachedTo = resultPoint;
        _attachedFrom = output;

        _xend = resultPoint.getX();
        _yend = resultPoint.getY();
        resultPoint.addAttachment(this);

        invalidatePath();
    }

    public InputConnector getAttachedTo() {
        return _attachedTo;
    }

    @Override
    public Moveable getMoveable() {
        return null;
    }

    public void unlink() {
        _attachedFrom.wireUnlinked(this);
        _attachedTo.wireUnlinked(this);
    }

    public Object query(Object lastValue) {
        return this._attachedFrom.query(lastValue);
    }

    private void invalidatePath() {
        path = null;
    }

    private Path getPath() {
        if (path == null) {
            path = new Path();

            // Bezier
            path.moveTo(_xinit, _yinit);
            path.cubicTo(
                    _xinit, _yinit + _pathRunWay,
                    _xend, _yend - _pathRunWay,
                    _xend, _yend);
        }
        return path;
    }

    @Override
    public boolean containsPoint(int x, int y) {
        return distanceToPoint(x, y) < MAX_DISTANCE_TO_CONTAINS;
    }

    private double distanceToPoint(int x, int y) {
        final int POINT_HALFSIDE = 25 / 2;
        RectF rectangle = new RectF(x - POINT_HALFSIDE, y - POINT_HALFSIDE,
                x + POINT_HALFSIDE, y + POINT_HALFSIDE);

        double sqDist1 = squareDistanceToSegment(_xinit, _yinit,
                _xinit, _yinit + _pathRunWay / 3,
                x, y
        );
        double sqDist2 = squareDistanceToSegment(
                _xinit, _yinit + _pathRunWay / 3,
                _xend, _yend - _pathRunWay / 3,
                x, y
        );
        double sqDist3 = squareDistanceToSegment(
                _xend, _yend - _pathRunWay / 3,
                _xend, _yend,
                x, y
        );

        return Math.sqrt(Math.min(sqDist1, Math.min(sqDist2, sqDist3)));
    }

    private double squareDistanceToSegment(int vx, int vy, int wx, int wy, int px, int py) {
        // Return minimum distance between line segment vw and point p
        // i.e. |w-v|^2 -  avoid a sqrt
        final double l2 = distance(vx, vy, wx, wy);
        if (l2 == 0.0) return distance(vx, vy, px, py);   // v == w case
        // Consider the line extending the segment, parameterized as v + t (w - v).
        // We find projection of point p onto the line.
        // It falls where t = [(p-v) . (w-v)] / |w-v|^2
        // We clamp t from [0,1] to handle points outside the segment vw.

        double pointInLine = ((px - vx) * (wx - vx) + (py - vy) * (wy - vy)) / l2;
        pointInLine = Math.max(0, Math.min(1, pointInLine));

        return distance(vx + pointInLine * (wx - vx),
                vy + pointInLine * (wy - vy),
                px, py
        );
    }

    private double distance(double vx, double vy, double px, double py) {
        return Math.pow(vx - px, 2) + Math.pow(vy - py, 2);
    }
}
