package com.programaker.bridge

import android.util.Log
import com.codigoparallevar.minicards.utils.Serializations
import com.google.gson.Gson
import okhttp3.*
import okio.ByteString
import org.json.JSONException
import org.json.JSONObject
import java.nio.charset.Charset
import java.util.concurrent.TimeUnit


class ProgramakerBridge(
        private val bridge_id: String,
        private val user_id: String,
        private val bridge_token: String,
        private val config: ProgramakerBridgeConfiguration,
        private val onReady: Runnable,
        private val onComplete: Runnable
) : WebSocketListener() {
    private var webSocket: WebSocket? = null
    private val utf8: Charset = Charset.forName("UTF-8")
    private val gson = Gson()

    private val PING_PERIOD_MILLIS: Long = 15000
    private val apiRoot: String = "wss://programaker.com/api"
    private val LogTag = "ProgramakerBridge"

    // Connection establishment
    fun run() {
        val client: OkHttpClient = OkHttpClient.Builder()
                .pingInterval(PING_PERIOD_MILLIS, TimeUnit.MILLISECONDS)
                .readTimeout(0, TimeUnit.MILLISECONDS)
                .build()

        val request: Request = Request.Builder()
                .url(getBridgeControlUrl())
                .build()
        client.newWebSocket(request, this)

        // Trigger shutdown of the dispatcher's executor so this process can exit cleanly.
        client.dispatcher.executorService.shutdown()
    }

    private fun getBridgeControlUrl(): String {
        return "$apiRoot/v0/users/id/$user_id/bridges/id/$bridge_id/communication"
    }

    // Websocket management
    override fun onOpen(webSocket: WebSocket, response: Response) {
        this.webSocket = webSocket

        val auth = JSONObject(hashMapOf(
                "type" to "AUTHENTICATION",
                "value" to hashMapOf (
                        "token" to this.bridge_token
                )
        ) as Map<*, *>).toString()

        webSocket.send(auth)

        webSocket.send(config.serialize())
        onReady.run()
    }

    override fun onMessage(webSocket: WebSocket, bytes: ByteString) {
        onMessage(webSocket, bytes.string(utf8))
    }

    override fun onMessage(webSocket: WebSocket, text: String) {
        Log.d(LogTag, "Message: $text")
        val json = gson.fromJson(text, HashMap::class.java)
        if (json == null) {
            this.onFailure(webSocket, JSONException("Error decoding: $text"), null)
        }
        else {
            try {
                this.onCommand(webSocket, json)
            }
            catch (ex: Throwable ) {
                Log.e(LogTag, "Error on command handling: $ex", ex)
            }
        }
    }

    override fun onClosing(webSocket: WebSocket, code: Int, reason: String) {
        webSocket.close(1000, null)
        Log.i(LogTag, "Closing bridge socket {code=$code, reason=$reason}")
        onComplete.run()
    }

    override fun onFailure(webSocket: WebSocket, t: Throwable, response: Response?) {
        Log.e(LogTag, "Error: $t", t)
        onComplete.run()
    }

    // Protocol handling
    private fun onCommand(webSocket: WebSocket, json: Map<*, *>) {
        val type = json.get("type") as String
        val messageId = json.get("message_id") as String
        val value = json.get("value") as Map<*, *>
        var userId: String? = null
        var extraData: Map<*, *>? = null
        if (json.containsKey("user_id")) {
            userId = json.get("user_id") as String
        }
        if (json.containsKey("extra_data")) {
            extraData = json.get("extra_data") as Map<*, *>
        }

        when (type) {
            "GET_HOW_TO_SERVICE_REGISTRATION" -> handleGetServiceRegistrationInfo(webSocket, value, messageId, userId, extraData)
            "REGISTRATION" -> handlePerformRegistration(webSocket, value, messageId, userId, extraData)
            "FUNCTION_CALL" -> handleFunctionCall(webSocket, value, messageId, userId, extraData)
            else ->
            {
                Log.w(LogTag, "Unknown command type: $type")
            }
        }
    }

    private fun handleFunctionCall(webSocket: WebSocket, value: Map<*, *>, messageId: String, userId: String?, extraData: Map<*, *>?) {
        val functionName = value.get("function_name") as String
        val arguments = value.get("arguments") as List<*>

        try {
            var result = config.callFunction(functionName, arguments)
            result = Serializations.blindSerialize(result)

            webSocket.send(
                    JSONObject(
                            hashMapOf(
                                    "message_id" to messageId,
                                    "success" to true,
                                    "result" to result
                            ) as Map<*, *>
                    ).toString()
            )
        }
        catch (ex: Throwable) {
            Log.w(LogTag, "Error on bridge call to $functionName", ex)
            webSocket.send(
                    JSONObject(
                            hashMapOf(
                                    "message_id" to messageId,
                                    "success" to false
                            ) as Map<*, *>
                    ).toString()
            )
        }

    }

    private fun handleGetServiceRegistrationInfo(webSocket: WebSocket, value: Map<*, *>, messageId: String?, userId: String?, extraData: Map<*, *>?) {
        // Registration is automatic
        webSocket.send(
                JSONObject(
                        hashMapOf(
                                "message_id" to messageId,
                                "success" to true,
                                "result" to null
                        ) as Map<*, *>
                ).toString()
        )
    }

    private fun handlePerformRegistration(webSocket: WebSocket, value: Map<*, *>, messageId: String, userId: String?, extraData: Map<*, *>?) {
        // Registration is automatic
        if (userId == null) {
            throw Exception("No connection ID received")
        }

        webSocket.send(
                JSONObject(
                        hashMapOf(
                                "message_id" to messageId,
                                "success" to true
                        ) as Map<*, *>
                ).toString()
        )
        this.config.configManager.bridgeConnectionId = userId
    }

    fun stop() {
        webSocket?.close(1000, null)
    }

    fun sendSignal(key: String, content: JSONObject) {
        if (webSocket == null) {
            Log.w(LogTag, "Cannot send signal (key=$key) on closed channel")
            return
        }

        webSocket!!.send(JSONObject(hashMapOf(
                    "type" to "NOTIFICATION",
                    "key" to key,
                    "subkey" to JSONObject.NULL,
                    "to_user" to JSONObject.NULL,
                    "value" to content,
                    "content" to content
                )
        ).toString())
    }
}